
package BlowfishJ;

import java.io.*;
import java.lang.*;


/**
  * testing the Blowfish encryption algorithm classes
  * @author Markus Hahn <markus_hahn@gmx.net>
  * @version October 09, 2002
  */

public class BlowfishTest 
{
  // max. size of message to encrypt

  final static int MAX_MESS_SIZE = 64;


  // benchmark settings

  final static int TESTBUFSIZE = 100000;
  final static int TESTLOOPS   = 1000;

     
  // startup CBC IV

  final static long CBCIV_START = 0x0102030405060708L;

    
  /**
    * the application entry point
    * @param args (command line) parameters
    */
  public static void main(String args[])
  {
    System.out.print("\nstarting selftest...");

    if (!BlowfishECB.selfTest()) 
    {
      System.out.println(",selftest failed");
      return;
    }
    
    System.out.println(", passed.");

    // create our test key

    byte[] testkey = new byte[5];
    for (int nI = 0; nI < testkey.length; nI++)
    {
      testkey[nI] = (byte)(nI + 1);
    }
     
    // do the key setups and check for weaknesses
  
    System.out.print("setting up Blowfish keys...");
   
    BlowfishECB bfecb = new BlowfishECB(testkey);
    
    BlowfishCBC bfcbc = new BlowfishCBC(testkey, CBCIV_START);
    
    System.out.println(", done.");
    
    if (bfecb.weakKeyCheck()) System.out.println("ECB key is weak!");
    else                      System.out.println("ECB key OK");

    if (bfcbc.weakKeyCheck()) System.out.println("CBC key is weak!");
    else                      System.out.println("CBC key OK");
      
    // get a message

    System.out.print("something to encrypt please >");
    System.out.flush();

    byte[] tempbuf = new byte[MAX_MESS_SIZE];

    int nMsgSize = 0;

    try 
    { 
      nMsgSize = System.in.read(tempbuf) - 2; // (cut off the line break)
      byte[] cpyBuf = new byte[nMsgSize];
      System.arraycopy(tempbuf, 0, cpyBuf, 0, nMsgSize);
      tempbuf = cpyBuf;
    }
    catch (java.io.IOException ioe) 
    { 
      return; 
    }

    // align to the next 8 byte border

    byte[] msgbuf;

    int nRest = nMsgSize & 7;

    if (nRest != 0) 
    {
      msgbuf = new byte[(nMsgSize & (~7)) + 8];

      System.arraycopy(tempbuf, 0, msgbuf, 0, nMsgSize);

      for (int nI = nMsgSize; nI < msgbuf.length ; nI++) 
      {
        msgbuf[nI] = 0;
      }
 
      System.out.println("message with " + nMsgSize + " bytes aligned to " + 
                         msgbuf.length + " bytes");
    }
    else 
    {
      msgbuf = new byte[nMsgSize];

      System.arraycopy(tempbuf, 0, msgbuf, 0, nMsgSize);
    }

    System.out.println("aligned data : " + BinConverter.bytesToBinHex(msgbuf));

    // ECB encryption/decryption test...

    bfecb.encrypt(msgbuf);

    // show the result

    System.out.println("ECB encrypted: " + BinConverter.bytesToBinHex(msgbuf));

    bfecb.decrypt(msgbuf);

    System.out.println("ECB decrypted: >>>" + new String(msgbuf) + "<<<");


    // CBC encryption/decryption test...

    byte[] showIV = new byte[BlowfishCBC.BLOCKSIZE];

    bfcbc.getCBCIV(showIV);

    System.out.println("CBC IV: " + BinConverter.bytesToBinHex(showIV));

    bfcbc.encrypt(msgbuf);


    // show the result

    System.out.println("CBC encrypted: " + BinConverter.bytesToBinHex(msgbuf));

    bfcbc.setCBCIV(CBCIV_START);
    bfcbc.decrypt(msgbuf);

    System.out.println("CBC decrypted: >>>" + new String(msgbuf) + "<<<");


    // test official vectors ...
    
    int nI = 0, nJ;
    
    while (nI < TEST_DATA.length)
    {
      byte[] key    = new byte[8];    long lKey    = TEST_DATA[nI++];
      byte[] plain  = new byte[8];    long lPlain  = TEST_DATA[nI++];
      byte[] cipher = new byte[8];    long lCipher = TEST_DATA[nI++];

      for (nJ = 7; nJ >= 0; nJ--)
      {
        key   [nJ] = (byte)(lKey    & 0x0ff);  lKey    >>>= 8;    
        plain [nJ] = (byte)(lPlain  & 0x0ff);  lPlain  >>>= 8;  
        cipher[nJ] = (byte)(lCipher & 0x0ff);  lCipher >>>= 8;     
      }
            
      byte[] testBuf = new byte[8];

      bfecb = new BlowfishECB(key);

      bfecb.encrypt(plain, testBuf);

      for (nJ = 0; nJ < 8; nJ++)
      {
        if (testBuf[nJ] != cipher[nJ])
        {
          System.out.println("error on vector " + nI / 3);
        }
      }
    }
    
    System.out.println("tests done.");
    
    // benchmark...

    System.out.println("\nbenchmark is running, encrypting " + 
                       TESTBUFSIZE * TESTLOOPS * 8 + " bytes\nplease wait...");  

    long lTm = System.currentTimeMillis();

    long[] testbuf = new long[TESTBUFSIZE];

    for (nI = 0; nI < TESTLOOPS; nI++) 
    {
      bfecb.encrypt(testbuf);

      if ((nI % 0x0f) == 0) 
      {
         System.out.print("#");
         System.out.flush();
      }    
    }

    lTm = System.currentTimeMillis() - lTm;

    System.out.println();

    double dAmount = TESTBUFSIZE * TESTLOOPS * 8;
    double dTime = lTm;
    double fRate = (dAmount * 1000) / dTime;
    long lRate = (long) fRate;

    System.out.println("rate: " + lRate + " bytes/sec");

    
    // don't leave any traces in the memory

    bfecb.cleanUp();
    bfcbc.cleanUp();
  }


  // (the official test vectors from the Counterpane website)

  final static long[] TEST_DATA = 
  {
      0x0000000000000000L, 0x0000000000000000L, 0x4EF997456198DD78L,
      0xFFFFFFFFFFFFFFFFL, 0xFFFFFFFFFFFFFFFFL, 0x51866FD5B85ECB8AL,
      0x3000000000000000L, 0x1000000000000001L, 0x7D856F9A613063F2L,
      0x1111111111111111L, 0x1111111111111111L, 0x2466DD878B963C9DL,
      0x0123456789ABCDEFL, 0x1111111111111111L, 0x61F9C3802281B096L,
      0x1111111111111111L, 0x0123456789ABCDEFL, 0x7D0CC630AFDA1EC7L,
      0x0000000000000000L, 0x0000000000000000L, 0x4EF997456198DD78L,
      0xFEDCBA9876543210L, 0x0123456789ABCDEFL, 0x0ACEAB0FC6A0A28DL,
      0x7CA110454A1A6E57L, 0x01A1D6D039776742L, 0x59C68245EB05282BL,
      0x0131D9619DC1376EL, 0x5CD54CA83DEF57DAL, 0xB1B8CC0B250F09A0L,
      0x07A1133E4A0B2686L, 0x0248D43806F67172L, 0x1730E5778BEA1DA4L,
      0x3849674C2602319EL, 0x51454B582DDF440AL, 0xA25E7856CF2651EBL,
      0x04B915BA43FEB5B6L, 0x42FD443059577FA2L, 0x353882B109CE8F1AL,
      0x0113B970FD34F2CEL, 0x059B5E0851CF143AL, 0x48F4D0884C379918L,
      0x0170F175468FB5E6L, 0x0756D8E0774761D2L, 0x432193B78951FC98L,
      0x43297FAD38E373FEL, 0x762514B829BF486AL, 0x13F04154D69D1AE5L,
      0x07A7137045DA2A16L, 0x3BDD119049372802L, 0x2EEDDA93FFD39C79L,
      0x04689104C2FD3B2FL, 0x26955F6835AF609AL, 0xD887E0393C2DA6E3L,
      0x37D06BB516CB7546L, 0x164D5E404F275232L, 0x5F99D04F5B163969L,
      0x1F08260D1AC2465EL, 0x6B056E18759F5CCAL, 0x4A057A3B24D3977BL,
      0x584023641ABA6176L, 0x004BD6EF09176062L, 0x452031C1E4FADA8EL,
      0x025816164629B007L, 0x480D39006EE762F2L, 0x7555AE39F59B87BDL,
      0x49793EBC79B3258FL, 0x437540C8698F3CFAL, 0x53C55F9CB49FC019L,
      0x4FB05E1515AB73A7L, 0x072D43A077075292L, 0x7A8E7BFA937E89A3L,
      0x49E95D6D4CA229BFL, 0x02FE55778117F12AL, 0xCF9C5D7A4986ADB5L,
      0x018310DC409B26D6L, 0x1D9D5C5018F728C2L, 0xD1ABB290658BC778L,
      0x1C587F1C13924FEFL, 0x305532286D6F295AL, 0x55CB3774D13EF201L,
      0x0101010101010101L, 0x0123456789ABCDEFL, 0xFA34EC4847B268B2L,
      0x1F1F1F1F0E0E0E0EL, 0x0123456789ABCDEFL, 0xA790795108EA3CAEL,
      0xE0FEE0FEF1FEF1FEL, 0x0123456789ABCDEFL, 0xC39E072D9FAC631DL,
      0x0000000000000000L, 0xFFFFFFFFFFFFFFFFL, 0x014933E0CDAFF6E4L,
      0xFFFFFFFFFFFFFFFFL, 0x0000000000000000L, 0xF21E9A77B71C49BCL,
      0x0123456789ABCDEFL, 0x0000000000000000L, 0x245946885754369AL,
      0xFEDCBA9876543210L, 0xFFFFFFFFFFFFFFFFL, 0x6B5C5A9C5D9E0A5AL
  };
}

