/** PMSProtocol.java
 *
 * By: Ken Hampson -- November, 2002
 *
 * This file provides an implementation of the WTalkProtocol interface. It facilitates the communication between the client and server
 * in the form of constructing messages to be sent back to the server. It also serves to connect the GUI with actual back-end functionality.
 */


package wtalk.protocols.pms;

// various imports
import wtalk.RMsgHandler;
import wtalk.IUserList;
import java.io.*;
import java.net.*;
import java.util.*;

public class PMSProtocol implements wtalk.protocols.WTalkProtocol{
  // assorted constants
  static final int DEFAULT_PORT = 3000;
  static final int TIME_SIZE = 4;
  static final int DATA_BUFFER_SIZE_SIZE = 2;
  static final int IDLE_TIME_SIZE = 4;
  static final int AWAY_TIME_SIZE = 4;
  static final int VERSION = 1;
  public static final String PMS_NAME = "PMS";

  // Message type codes
  static final int SEND_MESSAGE = 1;
  static final int RESERVED_2 = 2;
  static final int QUERY_STATUS_INFO = 3;
  static final int RETURN_STATUS_INFO = 4;
  static final int SIGN_ON = 5;
  static final int SIGN_OFF = 6;
  static final int GET_AWAY_MESSAGE = 7;
  static final int RETURN_AWAY_MESSAGE = 8;
  static final int POST_AWAY_MESSAGE = 9;
  static final int BAN_USER = 10;
  static final int UNBAN_USER = 11;
  /* 12 and 13 are currently unused */
  static final int SET_BUDDY_LIST = 14;
  static final int ADD_BUDDY = 15;
  static final int REMOVE_BUDDY = 16;
  static final int REMOVE_AWAY_MESSAGE = 17;
  static final int RECEIVE_AWAY_MESSAGE = 18;

  protected Socket mSock = null;
  protected DataOutputStream mDos = null;
  protected DataInputStream mDis = null;
  private ReceivedListener mRl = null;

  private String mServer = "";
  private String mName = "";
  private String mPasswd = "";
  private int mPort = 0;

  // message handlers
  private RMsgHandler mDmh = null; // Default message handler
  private RMsgHandler mSph = null; // Private message handler (if !null);
  private RMsgHandler mSmh = null; // System message handler (if !null);
  private RMsgHandler mSah = null; // Auto message handler (if !null);

  // user list
  private IUserList mUL = null;
  private String m_ConnectionName = null;


  /**Write out the individual components of the message header.
   * @author Ken Hampson
   * @param dos A DataOutputStream mapped to the socket to which the message header is to be written
   * @param version The version of the PMS protocol being used
   * @param msgType A code identifying the type of message that is being written out
   * @param dataSize The size of the data contained in the message, including an extra byte for each string to account for the null terminator
   * @param time The time to use to stamp the message (in ms from the epoch)
   * @throws An IOException is thrown when there is a problem writing to the DataOutputStream.
   */
  public void writeMsgHeader(DataOutputStream dos, int version, int msgType, int dataSize, int time) throws IOException, NumberFormatException {
    // write out the protocol version
    dos.writeByte(version);

    // write out the type of message
    dos.writeByte(msgType);

    // write out the data buffer size
    dos.writeShort((short)dataSize);

    // write out the time the message was sent (high byte written first)
    dos.writeInt(time);

    dos.flush();
  }



  /** Constructor
   * @author Ken Hampson
   */
  public PMSProtocol(String name){ System.out.println("Got PMS..."); PMSInit(name);}

  /** Initializes the PMSProtocol object
   * @author Matt Wronka
   */
  private void PMSInit(String name) {
    m_ConnectionName = name;
    PMSInit();
  }

 /** Initializes the PMSProtocol object
   * @author Matt Wronka
   */
  private void PMSInit() {
    mRl = new ReceivedListener();
  }

  /**
   * Reinitalizes reader call backs on the ReceivedListener object.
   *
   * @author Matt Wronka
   */ 
  private void resetReader() {
    mRl.ReportTo(mDmh);
    mRl.ReportPrivateTo(mSph);
    mRl.ReportSystemTo(mSmh);
    mRl.ReportAutoTo(mSah);
    mRl.UserList(mUL);
  }


  /**
   * Returns a connection name for use with user lists. 
   *
   * @return  Returns the connection's name.
   */
  public String getConnectionName() {
	return m_ConnectionName;
  }

  //
  // Callbacks
  //
  /**
   * Implements the function defined in WTalkProtocol
   *
   * @param Class which implements wtalk.RMsgHandler
   * @see wtalk.protocols.WTalkProtocol
   */
  public void setReceivedMessageHandler(RMsgHandler r){
    mRl.ReportTo(r);
    mDmh=r;
  }

 /**
  * Implements the function defined in WTalkProtocol
  *
  * @param Class which implements wtalk.RMsgHandler
  * @see wtalk.protocols.WTalkProtocol
  */
  public void setChatMessageHandler(RMsgHandler r){
    mRl.ReportTo(r);
    if(mDmh!=null) mDmh =r;
  }

 /**
  * Implements the function defined in WTalkProtocol
  *
  * @param Class which implements wtalk.RMsgHandler
  * @see wtalk.protocols.WTalkProtocol
  */
  public void setPrivateMessageHandler(RMsgHandler r){
    mRl.ReportPrivateTo(r);
    mSph=r;
  }

 /**
  * Implements the function defined in WTalkProtocol
  *
  * @param Class which implements wtalk.RMsgHandler
  * @see wtalk.protocols.WTalkProtocol
  */
  public void setWallMessageHandler(RMsgHandler r){
    mRl.ReportWallTo(r);
    // set global
  }

 /**
  * Implements the function defined in WTalkProtocol
  *
  * @param Class which implements wtalk.RMsgHandler
  * @see wtalk.protocols.WTalkProtocol
  */
  public void setSystemMessageHandler(RMsgHandler r){
    mRl.ReportSystemTo(r);
    mSmh = r;
  }
  
 /**
  * Implements the function defined in WTalkProtocol
  *
  * @param Class which implements wtalk.RMsgHandler
  * @see wtalk.protocols.WTalkProtocol
  */
  public void setAutoMessageHandler(RMsgHandler r){
	mRl.ReportAutoTo(r);
	mSah=r;
  }

  /** Sets handler for manipulating the buddy list
   * @author Matt Wronka
   * @param ul is a class that implements wtalk.IUserList
   */
  public void setUserListHandler(wtalk.IUserList ul){
    if(ul!=null)System.err.println("[ok] Got Userlist handler");
    mUL = ul;
    mRl.UserList(ul);
  }

  //
  // Uncooked functions
  //
  // sends literal string
  /**
   * Dummy function needed to implement WTalkProtocol.
   *
   * @see wtalk.protocols.WTalkProtocol
   */
  public void sendRaw(String msg){

  }


  //
  // Authentication/Login
  //


  // includes sign-on
  /** Connects to the server
   * @author Matt Wronka
   * @throws IOException if there is an I/O error
   */
  public void connect() throws IOException{
    if(mUL!=null) mUL.addConnection(getConnectionName());
    PMSInit();
    resetReader();

    mSock = new Socket(mServer, mPort==0? DEFAULT_PORT : mPort);
    mDos = new DataOutputStream(mSock.getOutputStream());
    mDis = new DataInputStream(mSock.getInputStream());

    // Move this until after authenticated ...
    if(mUL!=null){
      mRl.setConnectionName(getConnectionName());
    }

    if(mUL==null)
      System.err.println("no userlist");

    mRl.ListenTo(mDis);
    mRl.start();

    sendSignOn(mDos, mName, mPasswd, VERSION);
  }

  /** Connects to the server
   * @param login String representing login (user name)
   * @throws IOException if there's an I/O error
   */
  public void connect(String login) throws IOException{
    connect(mServer, mPort, login);
  }

  /** Connects to the server
   * @param server String representing host to connect to
   * @param port int representing port to connect to
   * @throws IOException if there's an I/O error
   */
  public void connect(String server, int port) throws IOException{
    mServer = server;
    mPort = port;
    connect();
  }

  /** Connect to the server
   * @param server String representing host to connect to
   * @param port int representing port to connect to
   * @param login String representing login (user name)
   * @throws IOException if there's an I/O error
   */
  public void connect(String server, int port, String login) throws IOException{
    mName = login;
    connect(server, port);
  }

  /**
   * Cleanup on disconnect.
   *
   * @see wtalk.protocols.WTalkProtocol
   */
  public void disconnect(String msg){
	if(mUL!=null){
		mUL.delConnection(getConnectionName());
	}
	if(isConnected()) mSock = null;
  }

  // Are we currently connected?
  /** Returns true if there is an active connection, false otherwise.
   *
   * @see wtalk.protocols.WTalkProtocol
   */
  public boolean isConnected(){
    return mSock != null && mSock.isConnected();
  }

  // sets handle
  /**
   * Sets the user's name for the active connection to <em>name</em>
   *
   * @param name User name to request.  Depending on the protocol
   *      it is not guaranteed that this request will be granted,
   *      or if it will even make since once connected.
   * @see wtalk.protocols.WTalkProtocol
   */
  public void setName(String name){
    mName = name;
  }

 /**
  * Returns the user's name (handle)
  *      understood by the current connection.
  *
  * @return The user's current handle on this connection.
  * @see wtalk.protocols.WTalkProtocol
  */
  public String getName(){
    return mName;
  }

 /**
  * Sets the password the connection will use to authenticate.
  * <p>
  * Note that this does not change the password to authenticate,
  * and no such method is provided under the PMS protocol. 
  *
  * @param passwd Password to use to authenticate. 
  * @see wtalk.protocols.WTalkProtocol
  */
  public void setPassword(String passwd){
    mPasswd = passwd;
  }

  /**
   * Dummy function needed to implement WTalkProtocol.
   *
   * @see wtalk.protocols.WTalkProtocol
   */
  public void changePassword(String oldpwd, String newpwd){
  }

  // sets server -- SHOULD NOT connect
 /**
  * Sets the server to use when <code>connect()</code> is called.
  * <p>
  * This function does not attempt to connect.
  *
  * @param server Host that <code>connect()</code> will connect to.
  * @see wtalk.protocols.WTalkProtocol
  */
  public void setServer(String server){
    mServer = server;
  }

  /** Returns name of current server.
   */
  public String getServer(){
    return mServer;
  }

  /** Sets the current server port
   * @author Matt Wronka
   * @param port The port the server should run on
   */
  public void setServerPort(int port){
    mPort=port;
  }

  /** Gets the current server port
   * @author Matt Wronka
   * @param port The port the server is running on
   */
  public int getServerPort(){
    return mPort;
  }

  //
  // IM functions
  //

  /**Sends a message from one user to another.
   * @author Ken Hampson
   * @param dos A DataOutputStream mapped to the socket to which the message header is to be written
   * @param to The name of the user to which the message should be sent
   * @param data The content of the message to be sent
   * @param version The version of the PMS protocol being used
   * @param recipients A vector of Strings containing the users to whom the message should be sent
   * @param chatID The name associated with the chat to which this message should be sent
   * @throws An IOException is thrown when a problems occurs trying to write to the DataOutputStream
   */
  public void sendMessage(DataOutputStream dos, String to, String data, int version, Vector recipients,
                          String chatID) throws IOException {

    int vectorSize = 0;					// the length (in bytes) of the vector's contents

    to = mName;

    // find out how big the vector is
    for(int i = 0; i < recipients.size(); i++){
      String currRecip = (String)recipients.get(i);
      vectorSize += currRecip.length() + 1;
    }

    // the size of the data associated with this message
    int dataSize = to.length() + data.length() + chatID.length() + vectorSize + recipients.size() + 3 + 5;

    // construct a calendar object with the current time
    GregorianCalendar cal = new GregorianCalendar();

    // extract the time in ms from the epoch from a Date object extracted from the calendar
    int time = (int)cal.getTime().getTime();

    // write out the message header
    writeMsgHeader(dos, version, SEND_MESSAGE, dataSize, time);

    // write out the length of to and the name itself
    dos.writeByte((byte)to.length() + 1);
    dos.writeBytes(to);
    dos.writeByte(0);
    dos.flush();

    // write out the number of recipients
    dos.writeByte((byte)recipients.size());

    // write out the size of each recipient and then the recipient itself
    for (int i = 0; i < recipients.size(); i++){
      String currRecip = (String)recipients.get(i);
      byte recipLength = (byte)currRecip.length();

      dos.writeByte(recipLength + 1);
      dos.writeBytes(currRecip);
      dos.writeByte(0);
      dos.flush();
    }

    // write out the size of the chat ID and the ID itself
    dos.writeByte((byte)chatID.length() + 1);
    dos.writeBytes(chatID);
    dos.writeByte(0);
    dos.flush();

    // write out the size of the message and then the message itself
    dos.writeShort((short)(data.length() + 1));
    dos.writeBytes(data);
    dos.writeByte(0);
    dos.flush();
  }

  /**Write out any "simple" message to the socket. This can be used to send the following message types:
   * - Query Status Info
   * - Get Away Message (here, set data to null)
   * - Post Away Message
   * - Ban user
   * - Unban User
   *
   * @author Ken Hampson
   * @param dos A DataOutputStream mapped to the socket to which the message header is to be written
   * @param data The data with which this command is dealing (varies depending on msgType)
   * @param version The version of the PMS protocol being used
   * @param msgType A code identifying the type of message that is being written out
   * @throws An IOException is thrown when a problems occurs trying to write to the DataOutputStream
   */
  public void sendSimpleMessage(DataOutputStream dos, String data, int version, int msgType) throws IOException{
    // construct a calendar object with the current time
    GregorianCalendar cal = new GregorianCalendar();

    // the size of the data itself plus one for a null term. plus another for the size itself
    int dataSize = data.length() + 1 + 1;

    // extract the time in ms from the epoch from a Date object extracted from the calendar
    int time = (int)cal.getTime().getTime();

    // write out the message header
    writeMsgHeader(dos, version, msgType, dataSize, time);

    // write out the length of the data and the data itself
    dos.writeByte((byte)data.length() + 1);
    dos.writeBytes(data);
    dos.writeByte(0);
    dos.flush();
  }

  /**Send updated status info to the server. Arguments which are not being updated should be null.
   *
   * @author Ken Hampson
   * @param dos A DataOutputStream mapped to the socket to which the message header is to be written
   * @param version The version of the PMS protocol being used
   * @param msgType A code identifying the type of message that is being written out
   * @param name The name of the user associated with this info
   * @param phoneNum A String representing the phone number of the user
   * @param email A String representing the e-mail address of the user
   * @param major A String reprsenting the name of the user's major
   * @param location A String indentifying the user's location
   * @param idle The user's idle time (in ms from the epoch)
   * @param away The user's away time (in ms from the epoch)
   * @throws An IOException is thrown when a problems occurs trying to write to the DataOutputStream
   */
  public void sendStatusInfo(DataOutputStream dos, int version, int msgType, String name, String phoneNum, String email,
                             String major, String location, int idle, int away) throws IOException
  {
    // the size of the data associated with this message
    int dataSize = name.length() + phoneNum.length() + email.length() + major.length() + location.length() +
                   IDLE_TIME_SIZE + AWAY_TIME_SIZE + 5;

    // construct a calendar object with the current time
    GregorianCalendar cal = new GregorianCalendar();

    // extract the time in ms from the epoch from a Date object extracted from the calendar
    int time = (int)cal.getTime().getTime();

    // write out the message header
    writeMsgHeader(dos, version, msgType, dataSize, time);

    // write out the length of the name and the name itself
    dos.writeByte(name.length() + 1);
    dos.writeBytes(name);
    dos.writeByte(0);
    dos.flush();

    // write out the length of the phoneNum and the phuneNum itself
    dos.writeByte(phoneNum.length() + 1);
    dos.writeBytes(phoneNum);
    dos.writeByte(0);
    dos.flush();

    // write out the length of the email address and the email address itself
    dos.writeByte(email.length() + 1);
    dos.writeBytes(email);
    dos.writeByte(0);
    dos.flush();

    // write out the length of the major and the major itself
    dos.writeByte(major.length() + 1);
    dos.writeBytes(major);
    dos.writeByte(0);
    dos.flush();

    // write out the length of the location and the location itself
    dos.writeByte(location.length() + 1);
    dos.writeBytes(location);
    dos.writeByte(0);
    dos.flush();

    // write out the idle time
    dos.writeInt(idle);
    dos.flush();

    // write out the away time
    dos.writeInt(away);
    dos.flush();
  }

  /**Send a message to the server with the user's name and password indicating they wish to logon to the system.
   * @author Ken Hampson
   * @param dos A DataOutputStream mapped to the socket to which the message header is to be written
   * @param name The name of the user trying to connect to the system
   * @param password The password of the user trying to connect to the system
   * @param version The version of the PMS protocol being used
   * @throws An IOException is thrown when a problems occurs trying to write to the DataOutputStream
   */
  protected  void sendSignOn(DataOutputStream dos, String name, String password, int version) throws IOException{
    // the size of the data associated with this message
    int dataSize = name.length() + password.length() + 2 + 2;

    // construct a calendar object with the current time
    GregorianCalendar cal = new GregorianCalendar();

    // extract the time in ms from the epoch from a Date object extracted from the calendar
    int time = (int)cal.getTime().getTime();

    // write out the message header
    writeMsgHeader(dos, version, SIGN_ON, dataSize, time);

    // write out the length of the user name and the name itself
    dos.writeByte((byte)name.length() + 1);
    dos.writeBytes(name);
    dos.writeByte(0);
    dos.flush();

    // write out the length of the password and the password itself
    dos.writeByte((byte)password.length() + 1);
    dos.writeBytes(password);
    dos.writeByte(0);
    dos.flush();
  }

  /**Send the server buddy list management commands. This can be used to send the following message types:
   * - Add Buddy (ADD_BUDDY)
   * - Remove Buddy (REMOVE_BUDDY)
   *
   * @author Ken Hampson
   * @param dos A DataOutputStream mapped to the socket to which the message header is to be written
   * @param grp The group with which the user is (to be) associated
   * @param user The name of the user on which to act
   * @param version The version of the PMS protocol being used
   * @param msgType A code identifying the type of message that is being written out
   * @throws An IOException is thrown when a problems occurs trying to write to the DataOutputStream
   */
  protected  void sendBuddyListCmd(DataOutputStream dos, String grp, String user, int version, int msgType) throws IOException{
    // construct a calendar object with the current time
    GregorianCalendar cal = new GregorianCalendar();

    // extract the time in ms from the epoch from a Date object extracted from the calendar

    int time = (int)cal.getTime().getTime();
    int dataSize = grp.length() + user.length() + 2 + 2;

    // write out the message header
    writeMsgHeader(dos, version, msgType, dataSize, time);

    // write out the length of the group name and the group itself
    dos.writeByte((byte)grp.length() + 1);
    dos.writeBytes(grp);
    dos.writeByte(0);

    // write out the length of the user name and the user itself
    dos.writeByte((byte)user.length() + 1);
    dos.writeBytes(user);
    dos.writeByte(0);
    dos.flush();
  }

  /** Function with which to test buddy list functionality. Adds the specified buddies (and groups implicitly, if they don't already exist). (Testing function only)
   * @author Ken Hampson
   */
  public void createTestBuddyList() {
    addBuddy("Chuck_Schuldiner", "Metal_Gods_From_Beyond_the_Grave");

    addBuddy("Kerry_King", "Slayer");
    addBuddy("Tom_Araya", "Slayer");

    addBuddy("Dave_Lombardo", "Drummers");
    addBuddy("Richard_Christi", "Drummers");
    addBuddy("Mike_Portnoy", "Drummers");
    addBuddy("John_Tempesta", "Drummers");

    addBuddy("Hanzi", "Singers");
    addBuddy("Russell_Allen", "Singers");
    addBuddy("James_LaBrie", "Singers");
    addBuddy("Matt_Barlow", "Singers");
    addBuddy("Tom_Araya", "Singers");

    addBuddy("John_Petrucci", "Guitarists");
    addBuddy("Jon_Schaffer", "Guitarists");
    addBuddy("Michael_Romeo", "Guitarists");
    addBuddy("Kerry_King", "Guitarists");

    addBuddy("Michael_Romeo", "Symphony_X");
    addBuddy("Russell_Allen", "Symphony_X");
  }

  /** Wrapper function for the sendMessage function called below. This is called from the GUI to send a message to someone.
   * @author Matt Wronka
   * @param to A String designating to whom the message should be sent
   * @param data A String containing the data (the message) to be sent
   */
  public void sendMessage(String to, String data){
    Vector v = new Vector();
    v.add(to);

    try{
      sendMessage(mDos, to, data, VERSION, v, "0");

      if(mSph!=null)
        mSph.handle(to, null, data);
      else if(mDmh!=null)
        mDmh.handle(to, null, data);
    }
    catch (Exception e){
    }

  }

  /**
   * Dummy function needed to implement WTalkProtocol.
   *
   * @see wtalk.protocols.WTalkProtocol
   */
  public void whois(String user){
  }

  /**
   * Adds a buddy to "group" on a list maintained on the server.
   * This list will be retrieved at logon.
   *
   * @param buddy Name of buddy to add.
   * @param group Name of group buddy is to be added to.
   */
  public void addBuddy(String buddy, String group) {
    try {
      sendBuddyListCmd(mDos, group, buddy, VERSION, ADD_BUDDY);
    }
    catch (IOException e){
    }

    mUL.addGroup(group, getConnectionName());
    mUL.addUser(buddy, group, getConnectionName());
  }


  /**
   * Removes a buddy from "group" on a list maintained on the server.
   * This list will be retrieved at logon.
   *
   * @param buddy Name of buddy to remove.
   * @param group Name of group buddy is to be removed from.
   */
  public void removeBuddy(String buddy, String group) {
    try {
      sendBuddyListCmd(mDos, group, buddy, VERSION, REMOVE_BUDDY);
    }
    catch (IOException e){
    }

    //System.err.println("Trying to remove " + buddy + " from " + group);

    mUL.delUser(buddy, group, getConnectionName());
  }

  /**
   * Sets or removes away message that the server will send to users
   * 	who try and message us.
   *
   * @param message The message to send when users contact us, or null
   *	to not send any message (removes away status).
   */
  public void setAway(String message) {
    try{
      if(message != null)
        sendSimpleMessage(mDos, message, VERSION, POST_AWAY_MESSAGE);
      else
        sendSimpleMessage(mDos, "", VERSION, REMOVE_AWAY_MESSAGE);
    }
    catch(IOException e){
    }
  }

  //
  // Chat functions
  //
  /**
   * Dummy function needed to implement WTalkProtocol.
   *
   * @see wtalk.protocols.WTalkProtocol
   */
  public void joinChat(String room){

  }

  /**
   * Dummy function needed to implement WTalkProtocol.
   *
   * @see wtalk.protocols.WTalkProtocol
   */
  public void partChat(String room){

  }

  /**
   * Dummy function needed to implement WTalkProtocol.
   *
   * @see wtalk.protocols.WTalkProtocol
   */
  public void chatMessage(String room, String data){

  }

  /**
   * Dummy function needed to implement WTalkProtocol.
   *
   * @see wtalk.protocols.WTalkProtocol
   */
  public void setTopic(String room, String topic){
    //createTestBuddyList();
  }

  /**
   * Dummy function needed to implement WTalkProtocol.
   *
   * @see wtalk.protocols.WTalkProtocol
   */
  // room may be null!
  public void chatList(String room){

  }

  /**
   * Dummy function needed to implement WTalkProtocol.
   *
   * @see wtalk.protocols.WTalkProtocol
   */
  public void sendWall(String msg){ }

  //
  // Admin functions
  //

  /**
   * Implements WTalkProtocol function.  
   * Sets banned status of <em>user</em> to the boolean value of
   * <em>banned</em>.
   *
   * @param user User whose banned state we are concerned.
   * @param banned If true, <em>user</em> is banned, otherwise
   *	banned status is removed.
   * @see wtalk.protocols.WTalkProtocol
   */
  public void setBan(String user, boolean banned) {
    // construct a calendar object with the current time
    GregorianCalendar cal = new GregorianCalendar();

    // extract the time in ms from the epoch from a Date object extracted from the calendar
    int time = (int)cal.getTime().getTime();
    try{
      // write out the message header
      writeMsgHeader(mDos, VERSION, banned ? BAN_USER : UNBAN_USER, user.length() + 2, time);

      // write out the length of the data and the data itself
      mDos.writeByte((byte)user.length() + 1);
      mDos.writeBytes(user);
      mDos.writeByte(0);
      mDos.flush();
    }
    catch(IOException e){
    }
  }
}
